
/*****
 If an update procedure was interrupted (eg. cable disconnected or power loss),
 then the device was just partially updated. It starts to loop in bootloader
 mode permanently.
 An indicator of this state is that 'lsusb' command lists a device record
 containing ID 04d8:fe52, which dissapears and reappears in 15-seconds itervals.

 The solution is to build this code in Recovery Mode and then run it.
 For building this code in the Recovery mode, uncomment the line below.
 (Afterwards, make sure to comment out the line again !)
*****/
//#define RecoveryMode	true



#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>

/* libudev-dev is required. You can inistall it by executing: sudo apt-get install libudev-dev */
#include <libudev.h>


#define CMD_ACK						0x0E
#define CMD_REBOOT					0x70
#define CMD_GET_CHIP_INFO			0x71


unsigned char hex2char(char hex);

bool usb_OpenComm(void);
void usb_CloseComm(void);
bool usb_SendData(int pLen);
bool usb_GetPath(bool BootloadedDevice);


char usb_DevicePath[32];
int usb_Handler;
unsigned char usb_Buffer[64];

struct timespec SleepVal;


int main(int argc, char **argv) {
	unsigned char line[80];
	FILE *fh;
	size_t lineLen;
	unsigned char uC;
	unsigned char Board;
	unsigned char BootloaderVer;
	int failCount;
	int i;

	printf("\n FF3x Family Update Utility\n");
	printf(" (c)2014-2015 FLYFISH TECHNOLOGIES d.o.o.\n\n");
	if (argc != 2) {
		printf(" Syntax:\n    FF3x-update <filename>\n\n");
		return EXIT_FAILURE; }

#ifndef RecoveryMode

	if(usb_GetPath(false) == false) {
		printf("*** Error locating FF3x USB device, check cable. Unable to continue, exiting...\n");
		return EXIT_FAILURE; }
	if(usb_OpenComm() == false) {
		printf("*** Error obtaining FF3x USB device for exclusive use. Make sure that you run this code with sufficient USB access permissions (eg. as root). Unable to continue, exiting...\n");
		return EXIT_FAILURE; }

	usb_Buffer[1] = CMD_GET_CHIP_INFO;
	if(usb_SendData(1) == false) {
		printf("*** Error obtaining FF3x silicon version. Unable to continue, exiting...\n");
		usb_CloseComm();
		return EXIT_FAILURE; }
	fh = fopen(argv[1], "r");
	line[0] = 0;
	while (line[0] != 0x0A) fread(line, 1, 1, fh);
	line[0] = 0;
	while (line[0] != 0x0A) fread(line, 1, 1, fh);
	line[0] = 0;
	while (line[0] != 0x0A) fread(line, 1, 1, fh);
	fread(line, 1, 6, fh);
	fclose(fh);

	if((usb_Buffer[1] == (hex2char(line[0]) * 16) + hex2char(line[1])) && (usb_Buffer[2] == (hex2char(line[2]) * 16) + hex2char(line[3]))) {
		printf("The chip's current silicon version is %d.%d\n", usb_Buffer[3], usb_Buffer[4]);
	} else {
		printf("\n*** The update file belongs to another member of FF3x family. Unable to continue, exiting...\n");
		usb_CloseComm();
		return EXIT_FAILURE; }

	printf("Rebooting the chip into bootloader mode (it takes 10 seconds).\n");
	usb_Buffer[1] = CMD_REBOOT;
	usb_SendData(1);
	usb_CloseComm();

	SleepVal.tv_sec = 10;
	SleepVal.tv_nsec = 0;
	nanosleep(&SleepVal, NULL);

	printf("Starting burn procedure (it takes few seconds).\n");
	if(usb_GetPath(true) == false) {
		printf("*** Error locating FF3x USB device (in bootloader mode). Unable to continue, exiting...\n");
		return EXIT_FAILURE; }

#else

	printf("Waiting for the device to loop into bootloader mode (timeout is 20 seconds).\n");
	failCount = 0;
	do {
		SleepVal.tv_sec = 1;
		SleepVal.tv_nsec = 0;
		nanosleep(&SleepVal, NULL);
		failCount++;
	} while ((usb_GetPath(true) == false) && (failCount < 20));
	if(failCount == 20) {
		printf("*** Timeout. The code was built with Recovery Mode option set - please verify that it is needed. Unable to continue, exiting...\n");
		return EXIT_FAILURE; }
	printf("Starting burn procedure (it takes few seconds).\n");

#endif

	if(usb_OpenComm() == false) {
		printf("*** Error obtaining FF3x USB device (in bootloader mode) for exclusive use. Make sure that you run this code with sufficient USB access permissions (eg. as root). Unable to continue, exiting...\n");
		return EXIT_FAILURE; }

	fh = fopen(argv[1], "r");
	line[0] = 0;
	while (line[0] != 0x0A) fread(line, 1, 1, fh);
	line[0] = 0;
	while (line[0] != 0x0A) fread(line, 1, 1, fh);
	line[0] = 0;
	while (line[0] != 0x0A) fread(line, 1, 1, fh);

	fread(line, 1, 6, fh);
	uC = (hex2char(line[0]) * 16) + hex2char(line[1]);
	Board = (hex2char(line[2]) * 16) + hex2char(line[3]);
	BootloaderVer = (hex2char(line[4]) * 16) + hex2char(line[5]);

	lineLen = fread(line, 1, 80, fh);
	for (i=0; i<40; i++) {
		usb_Buffer[i+1] = (hex2char(line[i*2]) * 16) + hex2char(line[(i*2)+1]); }
	if (usb_SendData(41) != true) {
		fclose(fh);
		usb_CloseComm();
		printf("*** Error communicating with the Device. Unable to continue, exiting...\n");
		return EXIT_FAILURE; }
	if (usb_Buffer[0] == CMD_GET_CHIP_INFO) {
		if((usb_Buffer[1] != uC) || (usb_Buffer[2] != Board) || (usb_Buffer[3] != BootloaderVer)) {
			fclose(fh);
			usb_CloseComm();
			printf("\n*** The update file does not belong to the Device detected. Unable to continue, exiting...\n");
			return EXIT_FAILURE; }
	} else {
		printf("\n*** Error detecting the Device - improper state. Start the procedure again. Unable to continue, exiting...\n");
		fclose(fh);
		usb_CloseComm();
		return EXIT_FAILURE; }

	failCount = 0;
	lineLen = fread(line, 1, 80, fh);
	while (lineLen == 80) {
		for (i=0; i<40; i++) {
			usb_Buffer[i+1] = (hex2char(line[i*2]) * 16) + hex2char(line[(i*2)+1]); }
		if (usb_SendData(41) == true) {
			if (usb_Buffer[0] == CMD_ACK) {
				failCount = 0;
				lineLen = fread(line, 1, 80, fh);
			} else {
				failCount++;
				if(failCount > 5) {
					fclose(fh);
					usb_CloseComm();
					printf("\n*** Error flashing the Device - too many failed retries. Start the procedure again. Unable to continue, exiting...\n");
					return EXIT_FAILURE; }	}
		} else {
			printf("\n*** Error flashing the Device - improper state. Start the procedure again. Unable to continue, exiting...\n");
			fclose(fh);
			usb_CloseComm();
			return EXIT_FAILURE; }	}

	usb_CloseComm();

	if (ferror(fh)) {
		printf("\n*** Error reading the input file. Please contact the vendor. Exiting...\n");
		fclose(fh);
		return EXIT_FAILURE; }

	fclose(fh);
	printf("Burn steps completed... rebooting the chip (it takes 10 seconds).\n");
	SleepVal.tv_sec = 10;
	SleepVal.tv_nsec = 0;
	nanosleep(&SleepVal, NULL);

	if(usb_GetPath(false) == false) {
		printf("*** Error locating FF3x USB device, check cable. Unable to continue, exiting...\n");
		return EXIT_FAILURE; }
	if(usb_OpenComm() == false) {
		printf("*** Error obtaining FF3x USB device for exclusive use. Make sure that you run this code with sufficient USB access permissions (eg. as root). Unable to continue, exiting...\n");
		return EXIT_FAILURE; }

	usb_Buffer[1] = CMD_GET_CHIP_INFO;
	if(usb_SendData(1) == false) {
		printf("*** Error obtaining FF3x silicon version. Unable to continue, exiting...\n");
		usb_CloseComm();
		return EXIT_FAILURE; }
	if((usb_Buffer[1] != uC) || (usb_Buffer[2] != Board)) {
		printf("\n*** Improper response (got from another member of FF3x family). Unable to continue, exiting...\n");
		usb_CloseComm();
		return EXIT_FAILURE; }

	printf("\nOK, procedure successfully completed, the chip's new silicon version is %d.%d\n\n", usb_Buffer[3], usb_Buffer[4]);

	usb_CloseComm();
	return EXIT_SUCCESS;
}


unsigned char hex2char(char hex) {
	if ((hex >= '0') && (hex <= '9')) {
		return (unsigned char)(hex - '0'); }
	if ((hex >= 'A') && (hex <= 'F')) {
		return (unsigned char)(hex - 'A' + 10); }
	return 0;
}



bool usb_OpenComm() {
	usb_Handler = open(usb_DevicePath, O_RDWR);
	if (usb_Handler < 0) {
		fprintf(stderr, "Unable to open USB device.\n");
		return false; }
	return true;
}

void usb_CloseComm() {
	close(usb_Handler);
}

bool usb_SendData(int pLen) {
	int Count;
	usb_Buffer[0] = 0x00;
	Count = write(usb_Handler, usb_Buffer, pLen+1);
	if (Count < 0) {
		return false; }
	Count = read(usb_Handler, usb_Buffer, 64);
	if (Count < 0) {
		return false; }
	return true;
}


bool usb_GetPath(bool BootloadedDevice) {
	struct udev *udev;
	struct udev_enumerate *enumerate;
	struct udev_list_entry *devices, *dev_list_entry;
	struct udev_device *dev;

	usb_DevicePath[0] = 0x00;
	udev = udev_new();
	if (!udev) {
		return false; }
	enumerate = udev_enumerate_new(udev);
	udev_enumerate_add_match_subsystem(enumerate, "hidraw");
	udev_enumerate_scan_devices(enumerate);
	devices = udev_enumerate_get_list_entry(enumerate);
	udev_list_entry_foreach(dev_list_entry, devices) {
		const char *path;
		path = udev_list_entry_get_name(dev_list_entry);
		dev = udev_device_new_from_syspath(udev, path);
		strcpy(usb_DevicePath, udev_device_get_devnode(dev));
		dev = udev_device_get_parent_with_subsystem_devtype(dev, "usb", "usb_device");
		if (!dev) {
			return false; }
		if (BootloadedDevice == true) {
			if ((strcmp(udev_device_get_sysattr_value(dev,"idVendor"), "04d8") == 0) && (strcmp(udev_device_get_sysattr_value(dev,"idProduct"), "fe52") == 0)) {
				udev_device_unref(dev);
				udev_enumerate_unref(enumerate);
				udev_unref(udev);
				return true; }
		} else {
			if ((strcmp(udev_device_get_sysattr_value(dev,"idVendor"), "04d8") == 0) && (strcmp(udev_device_get_sysattr_value(dev,"idProduct"), "f8b9") == 0)) {
				udev_device_unref(dev);
				udev_enumerate_unref(enumerate);
				udev_unref(udev);
				return true; }	}
		udev_device_unref(dev); }
	udev_enumerate_unref(enumerate);
	udev_unref(udev);
	return false;
}
